@@ -3,7 +3,7 @@
  * Author: Igor Vlasenko <vlasenko@imath.kiev.ua>
  * Created: Fri Jul  1 20:11:51 2005
  *
- * $Id: exprpstr.h,v 1.2 2005/11/22 20:55:34 igor Exp $
+ * $Id$
  */
 
 #ifndef _EXPRPSTR_H
@@ -12,7 +12,7 @@
 static
 PSTRING double_to_pstring (double, char* buf, size_t bufsize);
 static
-PSTRING int_to_pstring (int, char* buf, size_t bufsize);
+PSTRING int_to_pstring (EXPR_int, char* buf, size_t bufsize);
 static
 int pstring_ge(PSTRING, PSTRING);
 static
